// Path Finding Library for PCMod (SERVERSIDE)

// Define the Library
PCMPF = {}
PCMPF.Version = 1

Msg( "=> PCMod Pathfinding Library Initialized <=\n" )
Msg( "=> Version: " .. tostring( PCMPF.Version ) .. " <=\n" )

// Define all the functions
function PCMPF_FindPath( source, destination )
	// FindPath: Finds a router from the source subnet router to the destination subnet router
	// Outputs: A table of subnets to follow source --> dest, OR false if no path
	
	PCMPF_Msg( "FindPath Started" )
	
	// Check that the source and destination are valid registered routers
	local SEnt = PCMPF_GetNodeEntity( source )
	local DEnt = PCMPF_GetNodeEntity( destination )
	if ((!SEnt) || (!SEnt:IsValid()) || (!DEnt) || (!DEnt:IsValid())) then
		PCMPF_Msg( "Invalid SEnt or DEnt!" )
		PCMPF_Msg( "(" .. tostring(source) .. "), (" .. tostring(destination) .. ")" )
		return false
	end
	
	// Start off the node list
	PCMPF.NodeList = {}
	PCMPF_RegisterNode( source, source )
	
	// Start searching, and keep going until there are no items left to search (we are building an exaustive list)
	local i = 0
	repeat
		i = i + 1
		PCMPF_Msg( "Run through " .. tostring(i) )
		PCMPF_RunThroughList( )
	until (PCMPF_OpenListCount() == 0)
	
	// See if the target is in the list
	local found = PCMPF_NodeInList( destination )
	
	// If we couldn't find a path, stop here
	if (!found) then PCMPF_Msg( "Open list empty!" ); return false end
	
	// Otherwise, build the path
	PCMPF_Msg( "Building path..." )
	local path = PCMPF_BuildPath( source, destination )
	return path
end

function PCMPF_NodeInList( node )
	// NodeInList: Returns whether the node is registered or not
	// Outputs: TRUE if found, FALSE if not
	for k, v in pairs( PCMPF.NodeList ) do
		if (v.ID == node) then return true end
	end
	return false
end

function PCMPF_Msg( txt )
	// Msg: Prints a message to the player if debug mode is on
	// Outputs: nothing
	if (PCMPF.Debug) then PCMPF.Ply:PrintMessage( HUD_PRINTTALK, txt ) end
end

function PCMPF_BuildPath( source, destination )
	// BuildPath: Build the path once it has been found
	// Outputs: Path as table
	// Work BACKWARDs from the destination node; add each node's parent to the path and move on to the parent node
	local node = destination
	local parent = 0
	local path = {}
	repeat
		table.insert( path, node )
		parent = PCMPF_GetNodeParent( node )
		if (parent == node) then break end
	until ((node == source) || (node == 0))
	return path
end

function PCMPF_OpenListCount()
	// OpenListCount: Counts all items that are still open
	// Outputs: The count
	local i = 0
	for k, v in pairs( PCMPF.NodeList ) do
		if (v.Status == "open") then i=i+1 end
	end
	return i
end

function PCMPF_SearchNode( node )
	// SearchNode: Adds all linked items of a node to the list with status 'open'
	// Outputs: TRUE if target found, FALSE if not
	local ent = PCMPF_GetNodeEntity( node )
	if ((!ent) || (!ent:IsValid())) then return end
	local tosearch = ent.Data[ ent:EntIndex() ].R_RS
	for k, v in pairs( tosearch ) do
		if (v) then
			local id = v[1].Data[ v[1]:EntIndex() ].RID
			PCMPF_RegisterNode( id, node )
		end
	end
end

function PCMPF_GetNodeParent( node )
	// GetNodeParent: Retrieve a node's parent from either the open or the close list
	// Returns: The node's parent or FALSE if the function failed
	for k, v in pairs( PCMPF.NodeList ) do
		if (v.ID == node) then return v.Parent end
	end
	return 0
end

function PCMPF_RunThroughList( )
	// RunThroughList: Goes through every item in the node list, searches them if they're open, and closes the node
	// Outputs: TRUE if target found, FALSE if not
	for k, v in pairs( PCMPF.NodeList ) do
		if (v.Status == "open") then
			PCMPF_CloseNode( v.ID )
			PCMPF_SearchNode( v.ID )
		end
	end
end

function PCMPF_GetNodeEntity( node )
	// GetNode: Returns the router entity that is the physical entity tied with the node
	// Outputs: Router entity
	local ent = PCMod.Routers[ node ]
	return ent
end

function PCMPF_RegisterNode( node, parent )
	// RegisterNode: Ensures the node is on the list with the correct parent.
	// Outputs: Nothing
	for k, v in pairs( PCMPF.NodeList ) do
		if (v.ID == node) then v.Parent = parent; return end
	end
	local nodeitem = {}
	nodeitem.ID = node
	nodeitem.Parent = parent
	nodeitem.Status = "open"
	table.insert( PCMPF.NodeList, nodeitem )
	PCMPF_Msg( "Node registered: " .. tostring( node ) .. " (" .. tostring( parent ) .. ")" )
end

function PCMPF_OpenNode( node )
	// OpenNode: Sets a node's status to open
	// Outputs: Nothing
	for k, v in pairs( PCMPF.NodeList ) do
		if (v.ID == node) then v.Status = "open"; return end
	end
end

function PCMPF_CloseNode( node )
	// CloseNode: Sets a node's status to closed
	// Outputs: Nothing
	for k, v in pairs( PCMPF.NodeList ) do
		if (v.ID == node) then v.Status = "closed"; return end
	end
end